package cz.mendelu.savic.vida_cs_server.controller;

import cz.mendelu.savic.vida_cs.model.GameProcess;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.SocketException;

import cz.mendelu.savic.vida_cs.model.IPAddressConfig;
import cz.mendelu.savic.vida_cs_server.model.GameServer;
import java.util.ArrayList;
import javafx.application.Platform;

/**
 * Client handler for communicating with client and sending following message signals:
 *
 * Client -> Server:
 *   JOIN X - I wanna join game with number X (or create if I'm first).
 *   LEAVE X - I wanna leave waiting queue for game with number X.
 *   START X - I want to launch game with number X.
 *   GET_INFO X - I wanna know how many players are connected to game X.
 *   FINISH X - I finished game with number X (either game finished, player left, or error occured)
 *   GET_PROJECTOR_INFO - I wanna know if the projector is being used.
 *
 * Server -> Client:
 *   ACCEPT X - I accept your request to join game X. (he will be accepted as long
 *              as he is playing the game) -> player goes to INGAME status.
 *   REJECT X - I reject your request to join game X.
 *   CONNECTED X Y - The game with number X has Y players connected.
 *   MESSAGE text - Client, please notify the user with this message.
 *   IN_PROGRESS X - Game with number X is in progress (already played).
 *   LAUNCH X P - Client, please start game X which you requested on port P.
 *   END X - Client, close game X (if you didn't do it yet), because it finished.
 *   PROJECTOR_IN_USE - Client, projector is being used right now.
 *   PROJECTOR_NOT_IN_USE - Client, projector is available. You can use it.
 *   KILL_YOURSELF - Client, close your application.
 *
 * Where X is number of the game and Y number of players.
 *
 */
public class ClientHandler extends Thread {

    private BufferedReader input;
    private PrintWriter output;
    private Server server;
    private Socket socket;
    private GameServer actualGame;
    private boolean projectorInUse;

    public ClientHandler(Server server, Socket socket) {
        try {
            this.server = server;
            this.socket = socket;
            this.actualGame = null;
            this.projectorInUse = false;
            this.input = new BufferedReader(
                    new InputStreamReader(socket.getInputStream()));
            this.output = new PrintWriter(
                    socket.getOutputStream(), true);
        } catch (IOException e) {
            server.sendToServerConsole(e);
        }
    }

    public PrintWriter getOutput() {
        return output;
    }

    public Socket getSocket() {
        return socket;
    }

    public void sendToAClient(String message) {
        output.println(message);
    }

    public GameServer getActualGame() {
        return actualGame;
    }

    public void setActualGame(GameServer actualGame) {
        this.actualGame = actualGame;
    }

    public void wait(int miliseconds) {
        try {
            Thread.sleep(miliseconds);
        } catch (InterruptedException e) {
            server.sendToServerConsole(e);
        }
    }

    public void setProjectorInUse(boolean projectorInUse) {
        this.projectorInUse = projectorInUse;
    }

    private ArrayList<ClientHandler> getConnectedClients(GameServer game) {
        return game.getConnectedClients();
    }

    private ArrayList<ClientHandler> getNotConnectedClients(GameServer game) {
        return server.getNotConnectedClients(game);
    }

    private int getNumberOfConnectedClients(int gameIndex) {
        return server.getNonRunningGame(gameIndex).getNumberOfConnectedClients();
    }

    private boolean addConnectedClient(int gameIndex) {
        return server.getNonRunningGame(gameIndex).addConnectedClient(this);
    }

    private boolean removeConnectedClient(GameServer game) {
        return game.removeConnectedClient(this);
    }

    private boolean isActualGameRunning() {
        if (actualGame == null) {
            return false;
        }
        return actualGame.isRunning();
    }

    public void serverGameProcessCleanUp() {
        server.sendToServerConsole("Server pro hru " + (actualGame.getNumber())
                + " (instance " + actualGame.getInstance()
                + ") byl ukonen -> ukonuji i klienty");
        endGame();
    }

    /**
     * This method ends game. This cleans up on server and on each client. It is supposed to run
     * just once for the game (not on every ClientHandler, just one)
     */
    public void endGame() {
        if (actualGame == null || !isActualGameRunning()) {
            return;
        }
        GameServer game = actualGame;

        // SERVER:
        // set game isRunning to false
        game.setRunning(false);
        server.removeLastNonRunningGameInstance(game.getIndex());

        if (game.getIfUsesProjector()) {
            server.sendToAllClients("PROJECTOR_NOT_IN_USE");
        }

        // CLIENTS:
        // tell clients to terminated their launched app
        server.sendToSelectedClients("END " + game.getIndex(), getConnectedClients(game));

        // SERVER:
        // clear connectedClients, set their actualGame to none and projectorInUse to false
        game.clearConnectedClients();

        Platform.runLater(() -> {
            server.getController().setProjectorWindowAlwaysOnTop(true);
        });

        // terminate server app:
        game.endProcess();

        server.sendToServerConsole("Hra slo " + game.getNumber()
                + " (instance " + game.getInstance() + ") skonila");

        // CLIENTS:
        // set connected players on clients to 0
        server.sendToAllClients("CONNECTED " + game.getIndex() + " "
                + server.getNonRunningGame(game.getIndex()).getNumberOfConnectedClients());
    }

    @Override
    public void run() {
        while (true) {
            try {
                String message = input.readLine();
                String[] messageParts = message.split(" ");

                if (message.startsWith("GET_PROJECTOR_INFO")) {
                    server.sendToAllClients(projectorInUse
                            ? "PROJECTOR_IN_USE"
                            : "PROJECTOR_NOT_IN_USE");
                    continue;
                }

                int gameIndex = Integer.valueOf(messageParts[1]);
                int gameNumber = gameIndex + 1;

                if (message.startsWith("JOIN")) {
                    if (actualGame == null && !isActualGameRunning()
                            && (!server.getNonRunningGame(gameIndex).getIfUsesProjector()
                            || !projectorInUse)
                            && addConnectedClient(gameIndex)) {
                        setActualGame(server.getNonRunningGame(gameIndex));
                        this.sendToAClient("ACCEPT " + gameIndex);
                        server.sendToServerConsole("Klient na portu "
                                + socket.getPort()
                                + " se pidal k ekn na hru " + gameNumber
                                + " (instance " + server.getNonRunningGame(gameIndex).getInstance()
                                + ") poet pipojench hr je nyn "
                                + (getNumberOfConnectedClients(gameIndex)));
                        server.sendToAllClients("CONNECTED " + gameIndex + " "
                                + getNumberOfConnectedClients(gameIndex));
                    }
                } else if (message.startsWith("LEAVE")) {
                    if (!isActualGameRunning() && actualGame != null
                            && removeConnectedClient(actualGame)) {
                        server.sendToServerConsole("Klient na portu "
                                + socket.getPort()
                                + " ji neek na hru " + gameNumber + " (instance "
                                + actualGame.getInstance()
                                + ") poet pipojench hr je nyn "
                                + getNumberOfConnectedClients(gameIndex) + ")");
                        server.sendToAllClients("CONNECTED " + gameIndex + " "
                                + getNumberOfConnectedClients(gameIndex));
                        actualGame = null;
                    }
                } else if (message.startsWith("START")) {
                    if (actualGame != null && !isActualGameRunning()
                            && actualGame.getIndex() == gameIndex
                            && (!actualGame.getIfUsesProjector() || !projectorInUse)) {
                        if (actualGame.getIfUsesProjector()) {
                            projectorInUse = true;
                            server.sendToSelectedClients("PROJECTOR_IN_USE",
                                    getNotConnectedClients(actualGame));
                        }

                        int gamePort = actualGame.createGamePort();
                        server.sendToServerConsole("Spoutm hru " + gameNumber
                                + " (instance " + actualGame.getInstance()
                                + ") na portu " + gamePort);
                        String arguments = String.valueOf(gamePort) + " "
                                + new IPAddressConfig().getIPAddress();

                        actualGame.setRunning(true);
                        //server.addGameInstance(actualGame);

                        actualGame.setServerProcess(new GameProcess(actualGame.getRunCommandServer(),
                                arguments, this::serverGameProcessCleanUp));

                        actualGame.startServerProcess();

                        wait(1000);
                        // launch client apps
                        server.sendToSelectedClients("LAUNCH " + gameIndex + " " + gamePort,
                                getConnectedClients(actualGame));

                        server.sendToAllClients("CONNECTED " + actualGame.getIndex() + " 0");
                        if (actualGame.getIfUsesProjector()) {
                            Platform.runLater(() -> {
                                server.getController().setProjectorWindowAlwaysOnTop(false);
                            });
                        } else {
                            Platform.runLater(() -> {
                                server.getController().setProjectorWindowAlwaysOnTop(true);
                            });
                        }

                    }
                } else if (message.startsWith("GET_INFO")) {
                    sendToAClient("CONNECTED " + gameIndex + " " + server.getNonRunningGame(gameIndex).getNumberOfConnectedClients());
                } else if (message.startsWith("FINISH")) {
                    server.sendToServerConsole("Klient na portu "
                            + socket.getPort()
                            + " ukonil hru " + gameNumber + " (instance "
                            + actualGame.getInstance()
                            + ") -> zavrm i zbyl klienty tto hry");
                    endGame();
                }
            } catch (SocketException e) {
                // client disconneted
                try {
                    server.sendToServerConsole("Klient na portu " + socket.getPort()
                            + " se odpojil");
                    if (isActualGameRunning()) {
                        endGame();
                    } else if (actualGame != null) {
                        removeConnectedClient(actualGame);
                        server.sendToAllClients("CONNECTED " + actualGame.getIndex() + " "
                                + getNumberOfConnectedClients(actualGame.getIndex()));
                    }
                    socket.close();
                    return;
                } catch (IOException ex) {
                    server.sendToServerConsole(e);
                }
                break;
            } catch (IOException e) {
                server.sendToServerConsole(e);
            } catch (NullPointerException e) {
                server.sendToServerConsole(e);
            } catch (Exception e) {
                server.sendToServerConsole(e);
            } finally {
            }
        }
    }
}
